#!/usr/bin/env python3
# -*- coding=utf-8 -*-


class CircularDefinitionError(ValueError):
    pass


class InvalidSourceError(ValueError):
    pass


class StructureNotFoundError(ValueError):
    pass
